# flake8: noqa
# forwarding import
from cekit.descriptor.base import Descriptor
from cekit.descriptor.resource import Resource
from cekit.descriptor.label import Label
from cekit.descriptor.port import Port
from cekit.descriptor.env import Env
from cekit.descriptor.volume import Volume
from cekit.descriptor.osbs import Osbs
from cekit.descriptor.run import Run
from cekit.descriptor.packages import Packages
from cekit.descriptor.packages import Repository
from cekit.descriptor.modules import Modules
from cekit.descriptor.execute import Execute
from cekit.descriptor.image import Image
from cekit.descriptor.module import Module
from cekit.descriptor.overrides import Overrides
